#if (MACHINE == X68000)
/*
 * This file contains a driver for the ADPCM on the Sharp X68000.
 *
 * The driver supports only write operation.
 * It uses message format m2:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * | DISK_WRITE | device  | proc nr |  bytes  |         | buf ptr |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	adpcm_task: main entry when system is brought up
 *
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"
#include "x6addr.h"
#include "x6pcm.h"
#include "x6ppi.h"
#include "x6opm.h"
#include "x6dma.h"

#define	NR_PAN	1
#define TRACE(x)	/* x */

PRIVATE message mess;			/* message buffer for in and out */

PRIVATE void adpcminit();
PRIVATE int do_rdwt();
PRIVATE int do_ioctl();
PRIVATE void do_adpcm();

/*===========================================================================*
 *				adpcm_task				     *
 *===========================================================================*/
PUBLIC void adpcm_task()
{
  register int	r, minor, caller, procno;

 /*
  * Initialize ADPCM Controller and DMAC channel3
  */
  adpcminit();

  TRACE(printf("adpcm: task started.\n"));

  /*
   * The main loop of the adpcm task.
   * It waits for a message, carries it out, and sends a reply.
   */
  while (TRUE) {
	receive(ANY, &mess);
	TRACE(printf("adpcm: received %d from %d\n",mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case ADPCM_READ:
	    case ADPCM_WRITE:	r = do_rdwt(&mess);	break;
	    case ADPCM_IOCTL:	r = do_ioctl(&mess);	break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	if (caller >= 0){
	  mess.m_type = TASK_REPLY;	
	  mess.REP_PROC_NR = procno;
	  mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	  send(caller, &mess);	/* send reply to caller */
	}
  }
}

/*===========================================================================*
 *				adpcminit				     *
 *===========================================================================*/

PRIVATE void adpcminit()
{
  DMA3->csr = 0xff;
  DMA3->dcr = 0x80;
  DMA3->ocr = 0x32;
  DMA3->scr = 0x04;
  DMA3->mfc = 0x05;
  DMA3->dfc = 0x05;
  DMA3->bfc = 0x05;
  DMA3->niv = 0x76;
  DMA3->eiv = 0x77;
}

/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
  register struct proc	*rp;
  register		count, rw, minor;
  register phys_bytes	address;

  rw = mp->m_type;
  minor = mp->DEVICE;
  if (minor < 0 || minor >= NR_PAN)
	return(ENXIO);
  count = mp->COUNT;
  rp = proc_addr(mp->PROC_NR);
  address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) count);
  if (address == 0)
	return(E_BAD_ADDR);
  if (count <= 0)
	return(0);
  if (mp->m_type == DISK_WRITE)
	do_adpcm(address, (unsigned short) count);
  else
	return(EINVAL);
  return(count);
}


/*===========================================================================*
 *				do_ioctl				     * 
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;			/* pointer to read or write w_message */
{
  /* IOCTL is no support yet. */
  return(EINVAL);
}

/*===========================================================================*
 *				do_adpcm				     *
 *===========================================================================*/

PRIVATE void do_adpcm(addr, count)
phys_bytes	addr;
unsigned short	count;
{
  DMA3->ccr = 0x10;
  PCM->pcm_stat = 0x01;
  PPI->ppi_cwr = 0x07;
  PPI->ppi_cwr = 0x04;
  PPI->ppi_cwr = 0x02;
  PPI->ppi_cwr = 0x00;
  OPM->opm_regno = 0x1B;
  OPM->opm_data = 0x00;

  DMA3->csr = 0xff;
  DMA3->mtc = (unsigned short)count;
  DMA3->mar = (unsigned long)addr;
  DMA3->dar = &(PCM->pcm_data);
  DMA3->ccr = 0x88;
  PCM->pcm_stat = 0x02;
  receive(HARDWARE, &mess);
  PCM->pcm_data=0x88;
  PCM->pcm_stat = 0x01;
}

/*===========================================================================*
 *				pcmcint					     *
 *===========================================================================*/

PUBLIC void pcmcint()
{
  DMA3->csr = 0xff;
  interrupt(ADPCM);
}

/*===========================================================================*
 *				pcmeint					     *
 *===========================================================================*/

PUBLIC void pcmeint()
{
  printf ("dma error interrupt accepted.\n");
  printf ("status %02x %02x.\n", DMA3->csr, DMA3->cer);
  DMA3->ccr = 0x00;
  DMA3->cer = 0x00;
  DMA1->csr = 0xff;
}

#endif /* MACHINE == X68000 */
