#if (CHIP == M68000)
/*
 * This file contains special code dealing with the DMA interface
 * on the Sharp X68000.
 *
 * Although the DMA circuitry requires some special treatment
 * for the peripheral devices which make use of the ST's DMA feature
 * the DMA code constitutes not a usual Minix hardware process.
 * Nevertheless all devices using DMA need mutually exclusive
 * access and can follow some standard pattern which will be
 * provided in this file.
 *
 * The file contains the following entry points:
 *
 *	dmagrab:	ensure exclusive access to the DMA circuitry
 *	dmafree:	free exclusive access to the DMA circuitry
 *	dmaint:		DMA interrupt routine, switches to the current driver
 *	dmaaddr:	specify 24 bit RAM address
 *	dmardat:	set dma_mode and read word from dma_data
 *	dmawdat:	set dma_mode and write word to dma_data
 *	dmawcmd:	set dma_mode and write word to dma_data in one access
 *	dmacomm:	like dmawdat, but first toggle WRBIT
 */

#include "kernel.h"
#include <minix/com.h>
#include "proc.h"

#include "x6addr.h"
#include "x6dma.h"

#define	ASSERT(x)	if (!(x)) panic("dma: ASSERT(x) failed",NO_NUM);

PRIVATE	int	qhead;
PRIVATE	int	qlink[NR_TASKS];

#define	QNEXT(p)	qlink[(p) + NR_TASKS]

PUBLIC void dmainit();

/*===========================================================================*
 *				dmagrab					     *
 *===========================================================================*/
PUBLIC void dmagrab(p)
int p;
{
  message	m;
  register int	q;

  ASSERT(p >= -NR_TASKS && p < HARDWARE);
  ASSERT(QNEXT(p) == 0);
  if (q = qhead) {
	while (QNEXT(q))
		q = QNEXT(q);
	QNEXT(q) = p;
	receive(q, &m);
	ASSERT(qhead == p);
  } else
	qhead = p;
}

/*===========================================================================*
 *				dmafree					     *
 *===========================================================================*/
PUBLIC void dmafree(p)
int	p;
{
  message	m;

  ASSERT(qhead == p);
  qhead = QNEXT(p);
  QNEXT(p) = 0;
  if (qhead)
	send(qhead, &m);
}

/*===========================================================================*
 *				dmainit					     *
 *===========================================================================*/
PUBLIC void dmainit (p)
int	p;
{
  register struct dma *dma;
  static inited;

  if (!inited) {
	dmagrab(p);

       /*
	* Initialize DMAC channel 1
	*/
	dma = DMA1;
	dma->dcr = 0xc0;
	dma->scr = 0x04;
	dma->ccr = 0x00;
	dma->csr = 0xff;
	dma->cpr = 0x00;
	dma->mfc = 0x05;
	dma->dfc = 0x05;
	dma->bfc = 0x05;

	inited = 1;

	dmafree(p);
  }
}
#endif
#if 0
PRIVATE	int	(*xxxint)();
PUBLIC void dmagrab(p, func)
int (*func)();
  xxxint = func;		/* grab DMA interrupts */
  xxxint = (int(*)())0;		/* no more DMA interrupts */
/*===========================================================================*
 *				dmaint					     *
 *===========================================================================*/
PUBLIC void dmaint()
{
  if (xxxint)
	(*xxxint)();
  else
	printf("DMA interrupt discarded\n");
}
#endif
