#if (MACHINE == X68000)
/*
 * This file contains a driver for the Floppy Disk Controller (FDC)
 * on the Sharp X68000. It uses the NEC uPD72065 chip.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DISK_READ | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DISK_WRITE | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * | DISK_IOCTL | device  |         |         |         |         |
 * ----------------------------------------------------------------
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
 * ----------------------------------------------------------------
 *
 * The file contains two entry points:
 *
 *	floppy_task:	main entry when system is brought up
 *	fd_timer:	used to deselect drives, called from clock.c
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"

#include "x6addr.h"
#include "x6fdc.h"
#include "x6ioc.h"
#include "x6dma.h"

#define	ASSERT(x)	if (!(x)) panic("fd: ASSERT(x) failed",NO_NUM);
#define TRACE(x)	/* x */
#define DEBUG(x)	x

/* Parameters for the disk drive. */
#define NR_FLOPDRIVES	4	/* maximum number of drives */
#define NR_TYPES	3	/* number of diskette/drive combinations */

/*
 * X68000 floppy drives:
 * Three combination of diskette/drive are supported:
 *   # Drive  Diskette  Heads  Comment
 *   0   2HD    1200KB      2  1200KB
 *   1   2DD      720K      2  720KB 2DD (external drive only)
 *   2   2HD    1232KB      2  1232KB
 */

PRIVATE int sector_size[NR_TYPES] = {512, 512, 1024};	/* physical sector size */
PRIVATE int size_code[NR_TYPES]	= {2, 2, 3};		/* sector size code */
PRIVATE int nr_sectors[NR_TYPES] = {15, 9, 8};		/* number of sectors per track */
PRIVATE int nr_cylinders[NR_TYPES] = {80, 80, 77}; 	/* number of cylinders */
PRIVATE int nr_heads[NR_TYPES]	= {2, 2, 2};		/* tracks/cylinder */
PRIVATE int gap[NR_TYPES] = {0x1b, 0x1b, 0x35};		/* gap size */

#if 0
PRIVATE char spec[NR_TYPES] = {0xdf, 0xdf, 0xdf};	/* step rate, etc. */

PRIVATE int curspec = 0xdf;		/* latest spec sent to the controller */
#else
PRIVATE char spec[NR_TYPES] = {0xd0, 0xd0, 0xd0};	/* step rate, etc. */

PRIVATE int curspec = 0xd0;		/* latest spec sent to the controller */
#endif

PRIVATE int need_reset;		/* set to 1 when controller must be reset */
PRIVATE int res_status;
PRIVATE unsigned char results[MAX_RESULTS];
PRIVATE unsigned char dma_s, dma_e;	/* dma error status */
PRIVATE int curcyl[NR_FLOPDRIVES];	/* current cylinder (-1 if not calibrated) */


PRIVATE struct xfer {
	int	x_rw;		/* read or write */
	int	x_device;	/* device number */
	int	x_secnum;	/* current sector */
	phys_bytes x_address;	/* current physical address */
	int	x_count;	/* bytes still to transfer */
	int	x_drive;	/* drive number */
	int	x_type;		/* medium type */
	int	x_cylinder;	/* target cylinder */
	int	x_head;		/* target head */
	int	x_sector;	/* target sector */
	int	x_errors;	/* errors on current sector */
	int	x_cmd;		/* controller command */
} xfer;

PRIVATE message mess;		/* message buffer for in and out */

#define	DRIVE(d)	((d) & 0x07)	/* Drive No. */
#define	DTYPE(d)	(((d) >> 3) & 0x07)	/* diskette type */
#define FORMAT(d)	((d) & 0x40)

static int do_open();
static int do_rdwt();
static int do_xfer();
static int recalibrate();
static int seek();
static int transfer();
static void select();
static void deselect();
static void fdc_out();
static int fdc_results();
static void fdc_reset();
  
/*===========================================================================*
 *				floppy_task				     *
 *===========================================================================*/
PUBLIC void floppy_task()
{
  register r, drive, caller, procno;
  register struct dma *dma;

  IOC->int_mask = (IOC->int_stat & ~FDD_IEN) | FDC_IEN;
  res_status = fdc_results();
  fdc_reset ();

 /*
  * Initialize DMAC channel
  */
  dma = DMA0;
  dma->dcr = 0x80;
  dma->scr = 0x04;
  dma->csr = 0xff;
  dma->cpr = 0x00;
  dma->dar = (long) &FDC->data;	/* DMA window */
  dma->mfc = 0x05;
  dma->dfc = 0x05;
  dma->bfc = 0x05;
  dma->niv = 0x70;
  dma->eiv = 0x71;

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  TRACE(printf("fd: task started\n"));

  for (drive = 0; drive < NR_FLOPDRIVES; drive++)
	curcyl[drive] = -1;	/* uncalibrated */
  while (TRUE) {
	receive(ANY, &mess);
	TRACE(printf("fd: received %d from %d\n",mess.m_type,mess.m_source));
	ASSERT(mess.m_source >= 0);
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case DISK_READ:
	    case DISK_WRITE:	r = do_rdwt(&mess);	break;
	    case DISK_IOCTL:	r = do_ioctl(&mess);	break;
	    case SCATTERED_IO:	r = do_vrdwt(&mess, do_rdwt); break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}


/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
  register struct proc *rp;
  register struct xfer *xp;
  register nbytes;
  register int type;

  xp = &xfer;
  xp->x_rw = mp->m_type;
  xp->x_device = mp->DEVICE;
  xp->x_drive = DRIVE(xp->x_device);
  type = DTYPE(xp->x_device);
  xp->x_type = type;
  if (DRIVE(xp->x_drive) >= NR_FLOPDRIVES)
	return(EIO);
  if (type >= NR_TYPES)
	return(EIO);
  nbytes = mp->COUNT;
  xp->x_count = nbytes;
  if ((mp->POSITION % sector_size[type]) != 0)
	return(EINVAL);
  if ((nbytes % sector_size[type]) != 0)
	return(EINVAL);
  xp->x_secnum = (int)(mp->POSITION/sector_size[type]);
  rp = proc_addr(mp->PROC_NR);
  xp->x_address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) nbytes);
  if (xp->x_address == 0)
	return(EINVAL);
  if (nbytes == 0)
	return(0);
  rp->p_physio = 1;		/* disable (un)shadowing */
  xp->x_errors = 0;
  do_xfer();
  rp->p_physio = 0;		/* enable (un)shadowing */
  if (xp->x_count == nbytes)
	return(0);		/* nothing transferred */
  if (xp->x_count != 0)
	return(EIO);		/* partial transfer */
  return(nbytes);
}

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/
PRIVATE int do_xfer()
{
  register struct xfer *xp = &xfer;
  register int r = OK;
  int d, cylinder, type, errors;
  int restsec, restcount, skip;

  d = xp->x_drive;
  type = xp->x_type;
  restcount = xp->x_count;

  select(d, type);

  while (restcount > 0) {
	cylinder = xp->x_secnum / nr_sectors[type];
	xp->x_head = cylinder % nr_heads[type];
	xp->x_cylinder = cylinder / nr_heads[type];
	restsec = (xp->x_cylinder + 1) * nr_sectors[type] * nr_heads[type] - xp->x_secnum;
	if (restsec == 0)
		restsec = nr_sectors[type] * nr_heads[type];
	if ((xp->x_count / sector_size[type]) >restsec)
		xp->x_count = restsec * sector_size[type];
	restcount -= xp->x_count;
		skip = xp->x_count;

	if (xp->x_cylinder >= nr_cylinders[type])
		return(1);

	/* Set the stepping rate */
	if (curspec != spec[type]) {
		fdc_out(FDC_SPECIFY);
		fdc_out(curspec = spec[type]);
		fdc_out(SPEC2);
	}

	for (errors = 0; errors <= MAX_ERRORS; errors++) {
		if (need_reset)
			fdc_reset();

		if (curcyl[d] == -1)
			r = recalibrate(xp);

		/* If we are going to a new cylinder, perform a seek. */
		if (r == OK && xp->x_cylinder != curcyl[d])
			r = seek(xp);

		/* Perform the transfer. */
		if (r == OK)
			r = transfer(xp);

		if (r == OK) break;
	}
	if (need_reset){
		fdc_reset();
	}else{
		xp->x_address += skip;
		xp->x_secnum += skip / sector_size[type];
		xp->x_count = restcount;
	}
  }

  deselect();

  return(r);
}

/*===========================================================================*
 *				recalibrate				     * 
 *===========================================================================*/
PRIVATE int recalibrate(xp)
struct xfer *xp;
{
  int drive;

  drive = xp->x_drive;

  TRACE(printf("fd%d: recalibrate\n", drive));
  fdc_out(FDC_RECALIBRATE);	/* tell drive to recalibrate itself */
  fdc_out(drive);		/* specify drive */
  if (need_reset) return(ERR_SEEK);	/* don't wait if controller is sick */
  receive(HARDWARE, &mess);	/* wait for interrupt message */

  /* Determine if the recalibration succeeded. */
  if (res_status != OK ||		/* controller would not respond */
	   (results[ST0]&ST0_BITS) != SEEK_ST0 || results[ST_PCN] !=0){
		/* Recalibration failed.  FDC must be reset. */
		printf("fd%d: recalibration failed.\n", xp->x_device);
		need_reset = TRUE;
		return(ERR_RECALIBRATE);
  }
  curcyl[drive] = 0;
  return(OK);
}

/*===========================================================================*
 *				seek					     * 
 *===========================================================================*/
PRIVATE int seek(xp)
struct xfer *xp;
{
  int drive, cylinder;
  int r;

  drive = xp->x_drive;
  cylinder = xp->x_cylinder;

  /*
   * Issue a SEEK command on the indicated drive unless the arm is
   * already positioned on the correct cylinder.
   */

  TRACE(printf("fd%d: seek\n", drive));
  /* No.  Wrong cylinder.  Issue a SEEK and wait for interrupt. */
  fdc_out(FDC_SEEK);		/* start issuing the SEEK command */
  fdc_out(drive);
  fdc_out(cylinder);
  if (need_reset) return(ERR_SEEK);	/* if controller is sick, abort seek */
  receive(HARDWARE, &mess);

  /* Interrupt has been received.  Check drive status. */
  r = res_status;
  if ((results[ST0] & ST0_BITS) != SEEK_ST0) r = ERR_SEEK;
  if (results[ST1] != cylinder) r = ERR_SEEK;
  curcyl[drive] = (r == OK ? cylinder : -1);
  if (r != OK)  {
		printf("fd%d: seek failed.\n", xp->x_device);
		return(ERR_SEEK);
  }

  return(OK);
}

/*===========================================================================*
 *				transfer				     * 
 *===========================================================================*/
PRIVATE int transfer(xp)
register struct xfer *xp;
{
  int r, s;
  int sector, type;
  register struct dma *dma = DMA0;

  type = xp->x_type;

  /* The drive is now on the proper cylinder.  Read or write 1 block. */
  sector = xp->x_secnum % nr_sectors[type];
  sector++;	/* start numbering at 1 */

  TRACE(printf("fd%d: %s: secnum=%d,cylinder=%d,sector=%d,head=%d\n",
	xp->x_device, xp->x_rw == DISK_READ ? "read" : "write",
	xp->x_secnum, xp->x_cylinder, sector, xp->x_head));

  /* DMA transfer setup */
  dma_s = 0;
  dma_e = 0;
  dma->csr = 0xff;
  dma->mar = (long) xp->x_address;	/* DMA address setup */
  dma->mtc = xp->x_count;
  dma->ocr = xp->x_rw == DISK_READ ? 0xb2 : 0x32;	/* read / write */
  dma->ccr = 0x80;			/* start dma */

  /* The command is issued by outputting 9 bytes to the controller chip. */
  fdc_out(xp->x_rw == DISK_READ ? FDC_READ : FDC_WRITE);
  fdc_out((xp->x_head << 2) | xp->x_drive);
  fdc_out(xp->x_cylinder);
  fdc_out(xp->x_head);
  fdc_out(sector);
  fdc_out(size_code[type]);	/* sector size code */
  fdc_out(nr_sectors[type]);
  fdc_out(gap[type]);		/* sector gap */
  fdc_out(DTL);			/* data length */

  /* Block, waiting for disk interrupt. */
  if (need_reset) return(ERR_TRANSFER);	/* if controller is sick, abort op */
  receive(HARDWARE, &mess);		/* Wait for interrupt. */

  /* Get controller status and check for errors. */
  r = res_status;
  if (r != OK) return(r);
  if ((results[ST1] & BAD_SECTOR) || (results[ST2] & BAD_CYL))
	curcyl[xp->x_drive] = -1;
  if (results[ST1] & WRITE_PROTECT) {
	printf("Diskette in drive %d is write protected.\n", xp->x_drive);
	return(ERR_WR_PROTECT);
  }
  if ((results[ST0] & ST0_BITS) != TRANS_ST0 ||
      results[ST1] | results[ST2]) {
  	printf("fd%d: %s: secnum=%d,cylinder=%d,sector=%d,head=%d\n",
	xp->x_device, xp->x_rw == DISK_READ ? "read" : "write",
	xp->x_secnum, xp->x_cylinder, sector, xp->x_head);
	printf("status %02x %02x %02x \n", 
	        results[ST0], results[ST1], results[ST2]);
	return(ERR_TRANSFER);
  }

  xp->x_count = dma->mtc;

  return(OK);
}

/*==========================================================================*
 *				fdc_results				    *
 *==========================================================================*/
PRIVATE int fdc_results()
{
/* Extract results from the controller after an operation, then allow floppy
 * interrupts again.
 */

  int result_nr;
  register int retries;
  register int status;

  /* Extract bytes from FDC until it says it has no more.  The loop is
   * really an outer loop on result_nr and an inner loop on status.
   */
  result_nr = 0;
  retries = MAX_FDC_RETRY;
  while (TRUE) {
	/* Reading one byte is almost a mirror of fdc_out() - the DIRECTION
	 * bit must be set instead of clear, but the CTL_BUSY bit destroys
	 * the perfection of the mirror.
	 */
	status = FDC->stat & (MASTER | DIRECTION | CTL_BUSY);
	if (status == (MASTER | DIRECTION | CTL_BUSY)) {
		if (result_nr >= MAX_RESULTS) break;	/* too many results */
		results[result_nr++] = FDC->data;
		retries = MAX_FDC_RETRY;
		continue;
	}
	if (status == MASTER) {	/* all read */
		return(OK);	/* only good exit */
	}
	if (--retries == 0) break;	/* time out */
  }
  need_reset = TRUE;		/* controller chip must be reset */
  return(ERR_STATUS);
}

/*===========================================================================*
 *				fdc_out					     * 
 *===========================================================================*/
PRIVATE void fdc_out(val)
register int val;		/* write this byte to floppy disk controller */
{
/* Output a byte to the controller.  This is not entirely trivial, since you
 * can only write to it when it is listening, and it decides when to listen.
 * If the controller refuses to listen, the FDC chip is given a hard reset.
 */

  register int retries;

  if (need_reset) return;	/* if controller is not listening, return */

  /* It may take several tries to get the FDC to accept a command. */
  retries = MAX_FDC_RETRY;
  while ((FDC->stat & (MASTER | DIRECTION)) != (MASTER | 0))
	if (--retries == 0) {
  /* Controller is not listening.  Hit it over the head with a hammer. */
		need_reset = TRUE;
		return;
	}
  FDC->data = val;
}

/*===========================================================================*
 *				fdcint					     *
 *===========================================================================*/
PUBLIC void fdcint()
{
  if((FDC->stat & CTL_BUSY) == 0)
	fdc_out(FDC_SENSE);	/* issue SENSE command to request results */
  res_status = fdc_results();	/* get results of the FDC_??? command */
  if ((results[ST0] & 0x80) == 0)
	interrupt(FLOPPY);
}

/*===========================================================================*
 *			select,	deselect, fd_timer			     *
 *===========================================================================*/

PRIVATE int selected = -1;	/* drive/head currently selected */
PRIVATE int selectic = 30;	/* # of fd_timer calls to deselect */
PRIVATE int motoron;		/* for how long will motor run? */

PRIVATE void select(drive, type)
int drive;
int type;
{
  register s;

  TRACE(printf("fd: select (drive %d, type %d)\n", drive, type));

  selectic = -30;	/* error if not ready within 3 sec */
  if (drive != selected) {
	selected = drive;
	s = lock();
	FDC->dsel = MOTOR_ON | ((type == 1) << 4) | drive;
	restore (s);
	motoron = 5;		/* will run for more than 500 msec */
	receive(HARDWARE, &mess);
  }
}

PRIVATE void deselect()
{
  TRACE(printf("fd: deselect\n"));
  selectic = 30;	/* postpone for 30 * 100 msec == 3 sec */
}

PUBLIC void fd_timer()
{
/*
 * Perform two functions:
 * 1. Estimate when motor stops running
 * 2. Deselect drive(s) when all motors stopped spinning
 */
  register s, wrbit;

  if (selectic == 0)
	return;
  if (motoron)
	if (--motoron == 0)
		interrupt (FLOPPY);
  if (selectic < 0) {
	if (++selectic == 0) {
		s = lock();
		restore(s);
	}
	return;
  }
  if (--selectic != 0)
	return;
  selected = -1;
  s = lock();
  FDC->dsel = 0;
  restore(s);
}

#include <sgtty.h>

#define FD_EJECT       TIOCFLUSH	/* eject code */

/*===========================================================================*
 *				do_ioctl				     * 
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;			/* pointer to read or write w_message */
{
  int drive, request;
  
  drive = DRIVE(mp->DEVICE);
  if (drive < 0 || drive >= NR_FLOPDRIVES)
	return(EIO);
  	
  request = mp->COUNT;
  switch (request) {
  	case FD_EJECT:	FDC->dctl = EJECT | (1 << drive);
			FDC->dctl = 0;
  			break;
  	default:        return(EINVAL);
  }
  return OK;
}

/*===========================================================================*
 *				fdc_reset				     * 
 *===========================================================================*/
PRIVATE void fdc_reset()
{
/* Issue a reset to the controller.  This is done after any catastrophe,
 * like the controller refusing to respond.
 */

  int i;

  /* Disable interrupts and strobe reset bit low. */
  need_reset = FALSE;

  printf ("fdc reset\n");
  FDC->stat = FDC_RESET;	/* software reset */

  for (i=0; i<NR_DRIVES; i++)	/* Clear each drive. */
	curcyl[i] = -1;

  /* Tell FDC drive parameters. */
  fdc_out(FDC_SPECIFY);		/* specify some timing parameters */
  fdc_out(curspec);		/* step-rate and head-unload-time */
  fdc_out(SPEC2);		/* head-load-time and non-dma */
}
#endif
