#if (MACHINE == X68000)
/*
 * This file contains a driver for the SCSI Device Controller (SDC)
 * on the Sharp X68000.
 *
 * The driver supports two operations: read a block and write a block.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DISK_READ | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DISK_WRITE | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * | DISK_IOCTL | device  |         |         |         |         |
 * |--------------------------------------------------------------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	winchester_task: main entry when system is brought up
 *
 * Each controller support up to two drives.
 * Each physical drive contains up to 15 partitions, but for MINIX
 * each physical drive contains 16 minor devices, so that
 *	DEVICE = (drive << 4) | minor;
 * These 16 minor devices per drive are interpreted as follows:
 *	d+0:	whole disk	drive descriptor (sector 4)
 *	d+1:	partition 1	according to drive descriptor
 *	d+2:	partition 2	according to drive descriptor
 *	d+3:	partition 3	according to drive descriptor
 *	 :	    :
 *	d+15:	partition 15	according to drive descriptor
 *
 * Multiple controllers are not supported.
 * Multiple drives per controller are not yet tested.
 *
 * TODO:
 * - add more status checking of SDC and DMA
 * - find out which partitions are minix and which not
 * - obey bad sector info
 * - only access partitions with pi_id == "MIX" (prog to modify pi_id?)
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"

#include "x6addr.h"
#include "x6huscsi.h"
#include "x6dma.h"

#define TRACE(x)	/* x */
#define DEBUG(x)	x

/* Parameters for the disk drive. */
#define SECTOR_SIZE	512	/* physical sector size in bytes */
#define	MAX_MINOR	(NR_DRIVES<<4)

#define TIMEOUT		2000	/* SCSI bus timeout count */

PRIVATE unsigned char host_id;	/* SCSI id of X68000 */
PRIVATE struct pi pi[MAX_MINOR];	/* begin and size of any minor device */
PRIVATE struct hi hi[16];	/* buffer for drive descriptor */
PRIVATE unsigned char dma_s, dma_e;	/* dma error status */
PRIVATE message mess;		/* message buffer for in and out */
PRIVATE struct dma_chain array[NR_BUFS];
PRIVATE int chain;

PRIVATE int do_rdwt();
PRIVATE int do_xrdwt();
PRIVATE int do_xfer();
PRIVATE int get_p_info();
PRIVATE int do_scsi();
PRIVATE int scsi_selection();
PRIVATE void scsi_data_out();
PRIVATE void scsi_byte_out();
PRIVATE void scsi_data_in();
PRIVATE unsigned char scsi_byte_in();

/*===========================================================================*
 *				winchester_task				     *
 *===========================================================================*/
PUBLIC void winchester2_task()
{
  register int	r, drive, minor, caller, procno;
  register struct dma	*dma;

  dma = DMA2;

 /*
  * Initialize SCSI Controller
  */

  SDC->sc_mod = 0;
  host_id = 1 << (SDC->sc_dsw & 7);
  SDC->sc_dsw = 0;

 /*
  * Initialize DMAC channel
  */
  dma->dcr = 0xc0;
  dma->scr = 0x04;
  dma->csr = 0xff;
  dma->cpr = 0x00;
  dma->dar = (long) &SDC->sc_win;	/* DMA window */
  dma->mfc = 0x05;
  dma->dfc = 0x05;
  dma->bfc = 0x05;
  dma->niv = 0x74;
  dma->eiv = 0x75;

 /*
  * Initialize the partition description for all minor devices
  */
  for (drive = 0; drive < NR_DRIVES; drive++) {
	get_p_info ();
  }
  TRACE(
	for (r = 0; r < MAX_MINOR; r++)
		printf("hd%d: 1st=%D, tot=%D\n", r,
			pi[r].pi_start,
			pi[r].pi_size
		);
  );
  TRACE(printf("winchester: task started\n"));

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  while (TRUE) {
	receive(ANY, &mess);
	if (mess.m_source < 0)
		panic("disk task got message from ", mess.m_source);
	TRACE(printf("hd: received %d from %d\n",mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case DISK_READ:
	    case DISK_WRITE:	r = do_rdwt(&mess);	break;
	    case DISK_IOCTL:	r = do_ioctl(&mess);	break;
	    case SCATTERED_IO:	r = do_xrdwt(&mess);	break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}

/*==========================================================================*
 *				do_xrdwt				    *
 *==========================================================================*/
PUBLIC int do_xrdwt(m_ptr)
register message *m_ptr;	/* pointer to read or write message */
{
/* Fetch a vector of i/o requests.  Handle requests one at a time.  Return
 * status in the vector.
 */

  register struct iorequest_s *iop, *wiop, *eiop;
  static struct iorequest_s iovec[NR_BUFS];
  phys_bytes iovec_phys;
  unsigned nr_requests;
  int result;
  phys_bytes user_iovec_phys;
  message vmessage;
  int proc_nr;
  int device;
  register int bytes;
  register long position;

  nr_requests = m_ptr->COUNT;
  proc_nr = m_ptr->PROC_NR;
  device = m_ptr->DEVICE;
  if (nr_requests > sizeof iovec / sizeof iovec[0])
	panic("FS gave some driver too big an i/o vector", nr_requests);
  iovec_phys = umap(proc_ptr, D, (vir_bytes) iovec, (vir_bytes) sizeof iovec);
  user_iovec_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS,
			 (vir_bytes) (nr_requests * sizeof iovec[0]));
  if (user_iovec_phys == 0)
	panic("FS gave some driver bad i/o vector", (int) m_ptr->ADDRESS);
  phys_copy(user_iovec_phys, iovec_phys,
	    (phys_bytes) nr_requests * sizeof iovec[0]);

  iop = iovec;
  eiop = iop + nr_requests;
  while (iop < eiop) {
	vmessage.m_type = iop->io_request & ~OPTIONAL_IO;
	vmessage.DEVICE = device;
	vmessage.PROC_NR = proc_nr;
	vmessage.POSITION = position = iop->io_position;
	vmessage.ADDRESS = array[0].address = iop->io_buf;
	array[0].length = bytes = iop->io_nbytes;
	chain = 1;
	wiop = iop + 1;
	while (wiop < eiop &&
	    vmessage.m_type == (wiop->io_request & ~OPTIONAL_IO) &&
	    position + bytes == wiop->io_position) {
		bytes += (array[chain].length = wiop->io_nbytes);
		array[chain].address = wiop->io_buf;
		chain++;
		wiop++;
	}
	vmessage.COUNT = bytes;
	result = do_rdwt(&vmessage);
	chain = 0;
	if (result == 0) break;	/* EOF */
	if (result < 0) {
		iop->io_nbytes = result;
		if (iop->io_request & OPTIONAL_IO) break;  /* abort if opt */
	} else {
		while (iop < wiop) {
			if (iop->io_nbytes > result) {
				iop->io_nbytes -= result;
				break;
			} else {
				result -= iop->io_nbytes;
				iop->io_nbytes = 0;
				iop++;
			}
		}
	}
	iop = wiop;
  }

  phys_copy(iovec_phys, user_iovec_phys,
	    (phys_bytes) nr_requests * sizeof iovec[0]);
  return(OK);
}

/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
  register struct proc	*rp;
  register		r, errors, count, rw, drive, minor;
  register long		secnum, avail;
  register phys_bytes	address;

  rw = mp->m_type;
  minor = mp->DEVICE;
  drive = minor >> 4;
  if (drive < 0 || drive >= NR_DRIVES)
	return(EIO);
  if (pi[minor].pi_size == 0)
	return(EIO);
  if ((mp->POSITION % SECTOR_SIZE) != 0)
	return(EINVAL);
  secnum = (long)(mp->POSITION / SECTOR_SIZE);
  count = mp->COUNT;
  if ((count % SECTOR_SIZE) != 0)
	return(EINVAL);
  rp = proc_addr(mp->PROC_NR);
  address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) count);
  if (address == 0)
	return(EINVAL);
  count /= SECTOR_SIZE;
  TRACE(printf("hd%d: %s: sec=%D; cnt=%d\n",
	minor, rw == DISK_READ ? "read" : "write", secnum, count));
  avail = pi[minor].pi_size - secnum;
  if (avail <= 0)
	return(0);
  if (avail < count)
	count = avail;
  if (count <= 0)
	return(0);
  secnum += pi[minor].pi_start;
  rp->p_physio = 1;		/* disable (un)shadowing */
  r = do_xfer(drive, address, secnum, count, rw);
  rp->p_physio = 0;		/* enable (un)shadowing */
  if (r != OK)
	return(EIO);
  return(count * SECTOR_SIZE);
}

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/

PRIVATE int do_xfer(drive, address, sector, count, rw)
int		drive;
phys_bytes 	address;
long		sector;
int		count;
int		rw;
{
  register	r, s;
  unsigned char	command[6];

  /*
   * Carry out the transfer. All parameters have been checked and
   * are set up properly.
   *
   * Every single byte written to the hdc will cause an interrupt.
   * Thus disable interrupts while communicating with hdc. Ready test
   * will be done by busy waiting. Only for real hard disk operations
   * interrupts will be enabled.
   */

  dma_s = 0;
  dma_e = 0;

  command[0] = rw == DISK_READ ? SC_RD : SC_WR;
  command[1] = (sector >> 16) & 0x1f;
  command[2] = sector >> 8;
  command[3] = sector;
  command[4] = count;
  command[5] = 0;
  r = do_scsi (drive, command, 6, address, count * SECTOR_SIZE);

  return(r);
}

PRIVATE int do_scsi (target, command, clen, data, dlen)
int target;
unsigned char *command;
int clen;
char *data;
int dlen;
{
  int b, s, m;

  s = scsi_selection (target);
  if (s != 0)
	return s;

  do {
	b = SDC->sc_css & 0xfe;
	SDC->sc_trg = (b >> 2) & 7;	/* keep message, c/d, i/o */
	switch (b) {
	case PS_CMD:			/* command phase */
		scsi_data_out (command, clen);
		break;
	case PS_DIN:			/* data in phase */
		scsi_data_in (data, dlen);
		break;
	case PS_DOT:			/* data out phase */
		scsi_data_out (data, dlen);
		break;
	case PS_STT:			/* status phase */
		s = scsi_byte_in ();
		break;
	case PS_MES:			/* message phase */
		m = scsi_byte_in ();
		if (m == 1)
			m |= scsi_byte_in () << 8;
		break;
	}
  } while (b != PS_MES);		/* message phase */

  if ((s & 0x1e) != 0)
	printf ("status %02x message %02x\n", s, m);

  return s & 0x1e;
}

PRIVATE int scsi_selection (target)
int target;
{
  register int tm;
  register struct sdc *sdc;

  sdc = SDC;

  tm = sdc->sc_rpi & 1;			/* clear interrupt */
  if ((sdc->sc_css & 0x40) != 0)	/* bus busy ? */
	return ERROR;
  sdc->sc_odt = 1 << target;
  sdc->sc_trg = 0;
  sdc->sc_ini = 1;			/* assert bus */
  sdc->sc_ini = 5;			/* assert sel */
  for (tm += TIMEOUT; (sdc->sc_css & 0x40) != 0x40; tm--)
	if (tm == 0)
		return ERROR;
  sdc->sc_ini = 0;
  do {} while (!(sdc->sc_css & 0x20));	/* wait for req release */

  return OK;
}

PRIVATE void scsi_data_out (data, bytes)
unsigned char *data;
register int bytes;
{
  register struct dma *dma;
  register struct sdc *sdc;
  register unsigned char *css, *dat, *ini;

  dma = DMA2;
  sdc = SDC;
  css = &SDC->sc_css;
  dat = &SDC->sc_dat;
  ini = &SDC->sc_ini;

  *ini = 1;					/* assert bus */
  if (bytes < 512) {
	do {
		do {} while (!(*css & 0x20));	/* wait for req release */
		*dat = *data++;
		*ini = 0x11;			/* assert ack */
		*ini = 1;			/* release ack */
	} while (--bytes > 0);
  } else {
	sdc->sc_dsw = 1;
	dma->csr = 0xff;
	if (chain == 0) {
		dma->mar = (long) data;		/* DMA address setup */
		dma->mtc = bytes;
		dma->ocr = 0x32;
	} else {
		dma->bar = (long) array;	/* array address setup */
		dma->btc = chain;
		dma->ocr = 0x3a;
	}
	sdc->sc_mod = 2;			/* dma mode */
	dma->ccr = 0x88;			/* start dma */
	sdc->sc_sds = 0;

	receive(HARDWARE, &mess);		/* Wait for interrupt. */

	sdc->sc_mod = 0;			/* none dma mode */
	sdc->sc_dsw = 0;
  }
}

PRIVATE void scsi_byte_out (data)
unsigned char data;
{
  register struct sdc *sdc;

  sdc = SDC;

  sdc->sc_ini = 1;			/* assert bus */
  do {} while (!(sdc->sc_css & 0x20));	/* wait for req release */
  sdc->sc_odt = data;
  sdc->sc_ini = 0x11;			/* assert ack */
  sdc->sc_ini = 1;			/* release ack */
}

PRIVATE void scsi_data_in (data, bytes)
register unsigned char *data;
register int bytes;
{
  register struct dma *dma;
  register struct sdc *sdc;
  register unsigned char *css, *dat, *ini;

  dma = DMA2;
  sdc = SDC;
  css = &sdc->sc_css;
  dat = &sdc->sc_dat;
  ini = &sdc->sc_ini;

  *ini = 0;					/* release ack */
  if (bytes < 512) {
	do {
		do {} while (!(*css & 0x20));	/* wait for req release */
		*data++ = *dat;
		*ini = 0x10;			/* assert ack */
		*ini = 0;			/* release ack */
	} while (--bytes > 0);
  } else {
	sdc->sc_dsw = 1;
	dma->csr = 0xff;
	if (chain == 0) {
		dma->mar = (long) data;		/* DMA address setup */
		dma->mtc = bytes;
		dma->ocr = 0xb2;
	} else {
		dma->bar = (long) array;	/* array address setup */
		dma->btc = chain;
		dma->ocr = 0xba;
	}
	sdc->sc_mod = 2;			/* dma mode */
	dma->ccr = 0x88;			/* start dma */
	sdc->sc_sdi = 0;

	receive(HARDWARE, &mess);		/* Wait for interrupt. */

	sdc->sc_mod = 0;			/* none dma mode */
	sdc->sc_dsw = 0;
  }
}

PRIVATE unsigned char scsi_byte_in ()
{
  unsigned char data;
  register struct sdc *sdc;

  sdc = SDC;

  sdc->sc_ini = 0;			/* release ack */
  do {} while (!(sdc->sc_css & 0x20));	/* wait for req release */
  data = sdc->sc_dat;
  sdc->sc_ini = 0x10;			/* assert ack */
  sdc->sc_ini = 0;			/* release ack */
  return data;
}

#include <sgtty.h>

/*===========================================================================*
 *				do_ioctl				     * 
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;			/* pointer to read or write w_message */
{
  register	r, drive;

  drive = mp->DEVICE >> 4;
  if (drive < 0 || drive >= NR_DRIVES)
	return(EIO);
  	
  if (mp->COUNT == TIOCFLUSH)
	/*
	 * remake the partition description for the devices
	 */
	r = get_p_info (drive);
  else
	return(EINVAL);

  return(r);
}

PRIVATE int get_p_info (drive)
int drive;
{
  register	r, minor;

  minor = drive << 4;
  /* read sector 4 of the drive and copy the partition info */
  r = do_xfer(drive, (phys_bytes) hi, 4L, 1, DISK_READ);
  if (r != OK)
	return(EIO);
  for (r = 0; r < NPARTS; r++, minor++) {
	if (r == 0) {
		pi[minor].pi_size = *(unsigned long *) &hi[0].hi_act >> 1;
	} else {
		if (*(char *) hi[r].hi_name) {
			pi[minor].pi_name[0] = hi[r].hi_name[0];
			pi[minor].pi_name[1] = hi[r].hi_name[1];
			pi[minor].pi_start = hi[r].hi_start >> 1;
			pi[minor].pi_size = hi[r].hi_size >> 1;
		} else {
			pi[minor].pi_name[0] = 0;
			pi[minor].pi_name[1] = 0;
			pi[minor].pi_start = 0;
			pi[minor].pi_size = 0;
		}
	}
  }

  return OK;
}

/*===========================================================================*
 *				sdcint					     *
 *===========================================================================*/

PUBLIC void sdcint()
{
  DMA2->csr = 0xff;
  interrupt(WINCHESTER2);
}

PUBLIC void sdeint()
{
  dma_s = DMA2->csr;
  dma_e = DMA2->cer;
  printf ("dma error interrupt accepted.\n");
  printf ("status %02x %02x.\n", dma_s, dma_e);
  DMA2->csr = 0xff;
  interrupt(WINCHESTER2);
}
#endif 
