#if (MACHINE == X68000)
/*
 * This file contains a driver for the Hard Disk Controller (HDC)
 * on the Sharp X68000.
 *
 * The driver supports two operations: read a block and write a block.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DISK_READ | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DISK_WRITE | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * | DISK_IOCTL | device  |         |         |         |         |
 * |--------------------------------------------------------------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	winchester_task: main entry when system is brought up
 *
 * Each physical drive contains two of logical drives.
 * Each logical drive contains up to 15 partitions, so that
 *	DEVICE = (phys drive << 5) | (logical drive << 4) | minor;
 * These 16 minor devices per drive are interpreted as follows:
 *	d+0:	whole disk	drive descriptor (sector 4)
 *	d+1:	partition 1	according to drive descriptor
 *	d+2:	partition 2	according to drive descriptor
 *	d+3:	partition 3	according to drive descriptor
 *	 :	    :
 *	d+15:	partition 15	according to drive descriptor
 *
 * Multiple controllers are not supported.
 * Multiple drives per controller are not yet tested.
 *
 * TODO:
 * - add more status checking of HDC and DMA
 * - find out which partitions are minix and which not
 * - obey bad sector info
 * - only access partitions with pi_id == "MIX" (prog to modify pi_id?)
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"

#include "x6addr.h"
#include "x6sasi.h"
#include "x6dma.h"
#include "x6mfp.h"

#define TRACE(x)	/* x */
#define DEBUG(x)	x

/* Parameters for the disk drive. */
#define SECTOR_SIZE			256		/* physical sector size in bytes */
#define	MAX_MINOR			(1<<4)
#define NR_LOGICAL_DRIVES	(2)

/* begin and size of any minor device */
PRIVATE struct pi pi[NR_DRIVES*NR_LOGICAL_DRIVES*MAX_MINOR];
PRIVATE	struct hi hi[16];	/* buffer for drive descriptor */
PRIVATE char dma_s, dma_e;	/* dma error status */
PRIVATE message mess;		/* message buffer for in and out */

static int do_rdwt();
static int do_xfer();
static int do_test();
PRIVATE int do_ioctl();
static void cmd_phase_1();
static void cmd_phase_2();

#define WAIT_HDC_STAT(status) while (HDC->hd_stt != (status))

/*===========================================================================*
 *                              dma_setup                                    *
 *===========================================================================*/
PRIVATE void dma_setup()
{
	DMA1->niv = 0x72;
	DMA1->eiv = 0x73;
	DMA1->dar = (long) &HDC->hd_dat;
}

/*===========================================================================*
 *                             clear_flag                                    *
 *===========================================================================*/
PRIVATE void clear_flag()
{
	DMA1->csr = 0xff;
}

/*===========================================================================*
 *                       SASI send command                                   *
 *===========================================================================*/
PRIVATE void sasi_send_command(p1,p2,p3,p4,p5,p6)
unsigned char  p1, p2, p3, p4, p5, p6;
{
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p1;
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p2;
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p3;
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p4;
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p5;
	WAIT_HDC_STAT( PS_CMD | PS_REQ_BIT ); HDC->hd_dat = p6;
}

/*===========================================================================*
 *                         SASI drive select                                 *
 *===========================================================================*/
PRIVATE int sasi_select( phys )
unsigned int phys;
{
	int i, r = OK;

	WAIT_HDC_STAT( PS_BUS );
	HDC->hd_sel = 1 << phys;
	WAIT_HDC_STAT( PS_SEL );
/***
    hCu /usr/include/minix/config.h  NR_DRIVE Ŏw肳ꂽ
    菭ȂƂɂ́A܂ł[vĂ܂̂ŁAKȃ[vŕA
    悤ɂقǂł傤
    for( i = 0, r = ERROR ; i < 10000 ; i++ )
        if( HDC->hd_stt == PS_SEL ) { r = OK; break; }
***/
	HDC->hd_sre = 0;
	return(r);
}

/*===========================================================================*
 *                        DMA start                                          *
 *===========================================================================*/
PRIVATE void dma_start()
{
	DMA1->ccr = 0x88;
}

/*===========================================================================*
 *				winchester_task				     *
 *===========================================================================*/
PUBLIC void winchester_task()
{
	register int r, phys, logic, device, caller, procno;

   /*
	* Initialize DMAC channel
	*/
	dmainit (WINCHESTER);

	/*
     * Initialize the partition description for all minor devices
     */
	for (phys = 0; phys < NR_DRIVES; phys++)
	{
		for (logic = 0; logic < NR_LOGICAL_DRIVES; logic++ )
		{
			device = (phys << 5)|(logic << 4);

			/* read sector 0 of the drive and copy the partition info */
			if( do_test(device) != OK ) continue;
			if( do_xfer(device, (phys_bytes) hi, 4L, 1, DISK_READ) != OK ) continue;
			for (r = 0; r < NPARTS; r++, device++)
			{
				if (r == 0) pi[device].pi_size = hi[0].hi_name[1];
				else
				{
					if (*(char *) hi[r].hi_name)
					{
						pi[device].pi_name[0] = hi[r].hi_name[0];
						pi[device].pi_name[1] = hi[r].hi_name[1];
						pi[device].pi_start   = hi[r].hi_start & 0xffffff;
						pi[device].pi_size    = hi[r].hi_size;
					}
				}
			}
		}
	}
	TRACE(
	for (r = 0; r < MAX_MINOR; r++)
	    printf("hd%d: 1st=%D, tot=%D\n", r, pi[r].pi_start, pi[r].pi_size );
	);
	TRACE(printf("winchester: task started\n"));
	
	/*
     * The main loop of the disk task.
     * It waits for a message, carries it out, and sends a reply.
     */
	while (TRUE)
	{
		receive(ANY, &mess);
		if (mess.m_source < 0)
			panic("disk task got message from ", mess.m_source);
		TRACE(printf("hd: received %d from %d\n",mess.m_type,mess.m_source));
		caller = mess.m_source;
		procno = mess.PROC_NR;
		
		/* Now carry out the work. */
		switch (mess.m_type)
		{
		  case DISK_READ:
		  case DISK_WRITE:		r = do_rdwt(&mess);		break;
		  case DISK_IOCTL:		r = do_ioctl(&mess);	break;
		  case SCATTERED_IO:	r = do_vrdwt(&mess, do_rdwt); break;
		  default:				r = EINVAL;				break;
		}
		
		/* Finally, prepare and send the reply message. */
		mess.m_type      = TASK_REPLY;	
		mess.REP_PROC_NR = procno;
		mess.REP_STATUS  = r;	/* # of bytes transferred or error code */
		send(caller, &mess);	/* send reply to caller */
	}
}


/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
	register struct proc	*rp;
	register r, errors, count, rw, drive, phys, device;
	register long secnum, avail;
	register phys_bytes	address;

	rw = mp->m_type;
	device = mp->DEVICE;
	phys   = device >> 5;
	if( phys < 0 || phys >= NR_DRIVES) return(EIO);
	if(pi[device].pi_size == 0)        return(EIO);
	if ((mp->POSITION % SECTOR_SIZE) != 0) return(EINVAL);

	secnum = (long)(mp->POSITION / SECTOR_SIZE);
	count = mp->COUNT;
	if ((count % SECTOR_SIZE) != 0) return(EINVAL);

	rp = proc_addr(mp->PROC_NR);
	address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) count);
	if (address == 0) return(EINVAL);

	count /= SECTOR_SIZE;
	TRACE(printf("hd%d: %s: sec=%D; cnt=%d\n",
		   device, rw == DISK_READ ? "read" : "write", secnum, count));
	avail = pi[device].pi_size - secnum;
	if (avail <= 0) return(0);

	if (avail < count) count = avail;
	if (count <= 0) return(0);
	secnum += pi[device].pi_start;
	rp->p_physio = 1;	/* disable (un)shadowing */
	r = do_xfer(device, address, secnum, count, rw);
	rp->p_physio = 0;	/* enable (un)shadowing */
	if (r != OK) return(EIO);
	return(count * SECTOR_SIZE);
}

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/

PRIVATE int do_xfer(device, address, sector, count, rw)
int			device;
phys_bytes 	address;
long		sector;
int			count;
int			rw;
{
	register r, s;
	int phys, logic;
	unsigned char p1,p2,p3,p4,p5,p6;

	/*
     * Carry out the transfer. All parameters have been checked and
     * are set up properly.
     *
     * Every single byte written to the hdc will cause an interrupt.
     * Thus disable interrupts while communicating with hdc. Ready test
     * will be done by busy waiting. Only for real hard disk operations
     * interrupts will be enabled.
     */
	TRACE(printf("hd device:%d address:0x%X sector:%D count=%d cmd:%s\n",
		   device, address, sector, count, (rw==DISK_READ)?"READ":"WRITE")
	);

	/*** clear DMA status flag ***/
	dma_s = dma_e = 0;

	/*** compute physical and logical drive number of SASI ***/
	phys  = device >> 5;
	logic = (device >> 4) & 1;
  
	/*** make command line ***/
	if( rw == DISK_READ ) p1 = HD_RD;
	else                  p1 = HD_WR;
	p2 = ((logic << 5) & 0xe0) | ((sector >> 16) & 0x1f);
	p3 = (sector >> 8) &0xff;
	p4 = sector & 0xff;
	p5 = count;
	p6 = 0;

	/*** DMA settings ***/
	dmagrab(WINCHESTER);
	clear_flag();
	dma_setup();

	if( rw == DISK_READ ) DMA1->ocr = 0xb3;
	else                  DMA1->ocr = 0x33;
	DMA1->mar = address;
	DMA1->mtc = count * SECTOR_SIZE;

	/*** HDC: bus_free --> selection --> command ***/
	if( sasi_select( phys ) < 0 ) return ERROR;
	sasi_send_command( p1, p2, p3, p4, p5, p6 );

	/*** wait REQ signal ***/
	if (rw == DISK_READ)
		WAIT_HDC_STAT (PS_DRD|PS_REQ_BIT);
	else
		WAIT_HDC_STAT (PS_DWT|PS_REQ_BIT);

	/*** DMA start ***/
	dma_start();

	/*** wait transfer complete ***/
	receive(HARDWARE, &mess);
	dmafree(WINCHESTER);

	/*** HDC: get_status ***/
	WAIT_HDC_STAT (PS_STT|PS_REQ_BIT);
	s = HDC->hd_dat;

	/*** HDC: get_message ***/
	WAIT_HDC_STAT (PS_MES|PS_REQ_BIT);
	r = HDC->hd_dat;

	if( s & HDC_CC && dma_s & 0x10 )
	{
		printf("hd: %s: device=%d sector=%D status=0x%x\n",
                rw != DISK_READ ? "write" : "read", device, sector, s&0xFF);
		r = ERROR;
	} else r = 0;
	return(r);
}

/*===========================================================================*
 *				hdcint					     *
 *===========================================================================*/

PUBLIC void hdcint()
{
	DMA1->csr = 0xff;
	interrupt(WINCHESTER);
}

PUBLIC void hdeint()
{
	dma_s = DMA1->csr;
	dma_e = DMA1->cer;
	printf ("dma error interrupt accepted.\n");
	printf ("status %02x %02x.\n", dma_s, dma_e);
	DMA1->csr = 0xff;
	interrupt(WINCHESTER);
}

#include <sgtty.h>

/*===========================================================================*
 *				do_ioctl				     * 
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;			/* pointer to read or write w_message */
{
	register r, phys, logic, device;

	device = mp->DEVICE;
	if (device & 15) return(EIO);

	phys  = device >> 5;
	if (phys < 0 || phys >= NR_DRIVES) return(EIO);

	if (mp->COUNT == TIOCFLUSH)
	{
		/*
         * remake the partition description for the devices
         */
		/* read sector 0 of the drive and copy the partition info */
		if( (r = do_xfer(device, (phys_bytes) hi, 4L, 1, DISK_READ)) != OK ) return(EIO);

		for (r = 0; r < NPARTS; r++, device++)
		{
			if (r == 0) pi[device].pi_size = hi[0].hi_name[1];
			else
			{
				if (*(char *) hi[r].hi_name)
				{
					pi[device].pi_name[0] = hi[r].hi_name[0];
					pi[device].pi_name[1] = hi[r].hi_name[1];
					pi[device].pi_start   = hi[r].hi_start & 0xffffff;
					pi[device].pi_size    = hi[r].hi_size;
				}
				else
				{
					pi[device].pi_name[0] = 0;
					pi[device].pi_name[1] = 0;
					pi[device].pi_start   = 0;
					pi[device].pi_size    = 0;
				}
			}
		}
	} else return(EINVAL);

	return OK;
}

/*===========================================================================*
 *				do_test					     *
 *===========================================================================*/

PRIVATE int do_test(device)
int			device;
{
	unsigned char s, m;
	int phys, logic;
	unsigned char p1,p2,p3,p4,p5,p6;

	/*** compute physical and logical drive number of SASI ***/
	phys  = device >> 5;
	logic = (device >> 4) & 1;
  
	/*** make command line ***/
	p1 = 0x00;
	p2 = (logic << 5) & 0xe0;
	p3 = 0x00;
	p4 = 0x00;
	p5 = 0x00;
	p6 = 0x00;

	/*** HDC: bus_free --> selection --> command ***/
	if( sasi_select( phys ) < 0 ) return ERROR;
	sasi_send_command( p1, p2, p3, p4, p5, p6 );

	/*** HDC: get_status ***/
	WAIT_HDC_STAT (PS_STT|PS_REQ_BIT);
	s = HDC->hd_dat;

	/*** HDC: get_message ***/
	WAIT_HDC_STAT (PS_MES|PS_REQ_BIT);
	m = HDC->hd_dat;

	if ((s & ~0x0c) == 0) return OK;

	if( s != 0 ) return ERROR;
	else return OK;
}
#endif 
