#if (MACHINE == X68000)
/*
 * This file contains a driver for the SCSI Device Controller (SDC)
 * on the Sharp X68000.
 *
 * The driver supports two operations: read a block and write a block.
 * It accepts two messages, one for reading and one for writing,
 * both using message format m2 and with the same parameters:
 *
 *    m_type      DEVICE    PROC_NR     COUNT    POSITION  ADRRESS
 * ----------------------------------------------------------------
 * |  DISK_READ | device  | proc nr |  bytes  |  offset | buf ptr |
 * |------------+---------+---------+---------+---------+---------|
 * | DISK_WRITE | device  | proc nr |  bytes  |  offset | buf ptr |
 * |--------------------------------------------------------------|
 * | DISK_IOCTL | device  |         |         |         |         |
 * |--------------------------------------------------------------|
 * |SCATTERED_IO| device  | proc nr | requests|         | iov ptr |
 * ----------------------------------------------------------------
 *
 * The file contains one entry point:
 *
 *	winchester2_task: main entry when system is brought up
 *
 * Each SCSI disk support only one logical unit.
 * Each SCSI disk contains up to 15 partitions, but for MINIX
 * Each SCSI disk contains 16 minor devices, so that
 *	DEVICE = (device << 4) | minor;
 * These 16 minor devices per device are interpreted as follows:
 *	d+0:	whole disk	device descriptor (sector 4)
 *	d+1:	partition 1	according to device descriptor
 *	d+2:	partition 2	according to device descriptor
 *	d+3:	partition 3	according to device descriptor
 *	 :	    :
 *	d+15:	partition 15	according to device descriptor
 *
 * TODO:
 * - add more status checking of SDC and DMA
 */

#include "kernel.h"
#include <minix/callnr.h>
#include <minix/com.h>
#include "proc.h"

#include "x6addr.h"
#include "x6scsi.h"
#include "x6dma.h"

/*-----------------------------------------------------------------------------------
 * unfotunately if you still have some troubles about file system on scsi disk drive,
 * please try to define `DONT_USE_DMA' and rebuild kernel 
 *-----------------------------------------------------------------------------------*/
/* #define DONT_USE_DMA */

#define TRACE(x)	/* x */
#define DEBUG(x)	x

/* Parameters for the disk device. */
#define SECTOR_SIZE	512	/* physical sector size in bytes */
#define	MAX_DRIVES	8
#define	MAX_MINOR	(MAX_DRIVES<<4)

PRIVATE unsigned char b[SECTOR_SIZE];	/* for any purpose */
PRIVATE struct pi pi[MAX_MINOR];	/* begin and size of any minor device */
PRIVATE struct hi *hi = (struct hi *) b;/* buffer for device descriptor */
PRIVATE struct sdc *SDC;		/* SCSI controller address */
PRIVATE unsigned short timeout;		/* selection timeout value */
PRIVATE unsigned char dma_s, dma_e;	/* dma error status */
PRIVATE message mess;			/* message buffer for in and out */
PRIVATE unsigned char scsi_cmd[10];	/* SCSI command buffer */
PRIVATE unsigned char scsi_dat[12];	/* SCSI transfer buffer */

PRIVATE int do_rdwt();
PRIVATE int do_xfer();
PRIVATE int get_p_info();
PRIVATE int get_d_info();
PRIVATE int do_scsi();
PRIVATE int do_check();
PRIVATE int scsi_selection();
PRIVATE void scsi_data_out();
PRIVATE void scsi_byte_out();
PRIVATE void scsi_data_in();
PRIVATE unsigned char scsi_byte_in();

/*===========================================================================*
 *				winchester2_task			     *
 *===========================================================================*/
PUBLIC void winchester2_task()
{
  register int	r, target, minor, caller, procno;
  register struct dma	*dma;

 /*
  * Initialize DMAC channel
  */
  dmainit (WINCHESTER2);

  if (SCSI_VALID) {
       /*
	* Initialize SCSI Controller
	*/
	SDC = SCSI_IF ? EXT_SCSI : INT_SCSI;
	SDC->sc_sctl = 0x90;
	SDC->sc_sctl = 0x00;
	SDC->sc_sdgc = 0x00;
	SDC->sc_ints = 0xff;
	SDC->sc_bdid = HOST_ID;

       /*
	* Initialize the partition description for all minor devices
	*/
	for (target = 0; target < MAX_DRIVES; target++) {
		r = get_d_info (target);
		if (r != 0)
			continue;
		get_p_info (target);
		TRACE(
			for (r = 0; r < MAX_MINOR; r++)
				if (pi[r].pi_size == 0)
					continue;
				printf("sd%d%c: 1st=%D, tot=%D\n",
					target, "0123456789abcdef"[r],
					pi[r].pi_start,
					pi[r].pi_size
				);
		);
	}
  }
  TRACE(printf("winchester2: task started.\n"));

  /*
   * The main loop of the disk task.
   * It waits for a message, carries it out, and sends a reply.
   */
  while (TRUE) {
	receive(ANY, &mess);
	if (mess.m_source < 0)
		panic("disk task got message from ", mess.m_source);
	TRACE(printf("sd: received %d from %d\n",mess.m_type,mess.m_source));
	caller = mess.m_source;
	procno = mess.PROC_NR;

	/* Now carry out the work. */
	switch (mess.m_type) {
	    case DISK_READ:
	    case DISK_WRITE:	r = do_rdwt(&mess);	break;
	    case DISK_IOCTL:	r = do_ioctl(&mess);	break;
	    case SCATTERED_IO:	r = do_vrdwt(&mess, do_rdwt);	break;
	    default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	mess.m_type = TASK_REPLY;	
	mess.REP_PROC_NR = procno;
	mess.REP_STATUS = r;	/* # of bytes transferred or error code */
	send(caller, &mess);	/* send reply to caller */
  }
}

/*===========================================================================*
 *				do_rdwt					     *
 *===========================================================================*/
PRIVATE int do_rdwt(mp)
register message *mp;
{
  register struct proc	*rp;
  register		r, errors, count, rw, target, minor;
  register long		secnum, avail;
  register phys_bytes	address;

  register phys_bytes	address2;
  register long		secnum2;
  register int		i, stat;

  rw = mp->m_type;
  minor = mp->DEVICE;
  target = minor >> 4;
  if (target < 0 || target >= MAX_DRIVES)
	return(EIO);
  do {
	if (pi[target << 4].pi_size == 0) {
		if (!D_AVAL (target))
			return(EIO);
		r = get_d_info (target);
		if (r == 0)
			get_p_info (target);
	}
	if (pi[minor].pi_size == 0)
		return(EIO);
	if ((mp->POSITION % SECTOR_SIZE) != 0)
		return(EINVAL);
	secnum = (long)(mp->POSITION / SECTOR_SIZE);
	count = mp->COUNT;
	if ((count % SECTOR_SIZE) != 0)
		return(EINVAL);
	rp = proc_addr(mp->PROC_NR);
	address = umap(rp, D, (vir_bytes) mp->ADDRESS, (vir_bytes) count);
	if (address == 0)
		return(EINVAL);
	count /= SECTOR_SIZE;
	TRACE(printf("sd%d: %s: sec=%D; cnt=%d\n",
		minor, rw == DISK_READ ? "read" : "write", secnum, count));
	avail = pi[minor].pi_size - secnum;
	if (avail <= 0)
		return(0);
	if (avail < count)
		count = avail;
	if (count <= 0)
		return(0);
	secnum += pi[minor].pi_start;
	rp->p_physio = 1;		/* disable (un)shadowing */

	r = OK;
	for (i = 0; i < count; i++){
		address2 = address + SECTOR_SIZE * i;
		secnum2 = secnum + i;
		stat = do_xfer(target, address2, secnum2, 1, rw);
		switch (stat){
		case OK:
			break;
		case RETRY:
			if (r == OK){
				r = stat;
			}
			break;
		default:
			r = stat;
			break;
		}
	}

	rp->p_physio = 0;		/* enable (un)shadowing */
  } while (r == RETRY);
  if (r != OK)
	return(EIO);
  return(count * SECTOR_SIZE);
}

/*===========================================================================*
 *				do_xfer					     *
 *===========================================================================*/

PRIVATE int do_xfer(target, address, sector, count, rw)
int		target;
phys_bytes 	address;
long		sector;
int		count;
int		rw;
{
  register int	r, s;

  /*
   * Carry out the transfer. All parameters have been checked and
   * are set up properly.
   *
   * Every single byte written to the sdc will cause an interrupt.
   * Thus disable interrupts while communicating with sdc. Ready test
   * will be done by busy waiting. Only for real hard disk operations
   * interrupts will be enabled.
   */
  scsi_cmd[0] = rw == DISK_READ ? SC_RD : SC_WR;
  scsi_cmd[1] = (sector >> 16) & 0x1f;
  scsi_cmd[2] = sector >> 8;
  scsi_cmd[3] = sector;
  scsi_cmd[4] = count;
  scsi_cmd[5] = 0;

  dma_s = dma_e = 0;

  r = do_scsi (target, scsi_cmd, 6, address, count * SECTOR_SIZE);

  switch (r){
  case SD_CC:	/* Check Condition */
    s = do_check (target);
    break;
  case 0x08:	/* BUSY */
    s = RETRY;
    break;
  default:
    s = r;
    break;
  }

  return(s);
}

PRIVATE int do_check (target)
int		target;
{
  register int	r, s;

  scsi_cmd[0] = SC_RQS;
  scsi_cmd[4] = 4;
  scsi_cmd[1] = scsi_cmd[2] = 0;
  scsi_cmd[3] = scsi_cmd[5] = 0;
  r = do_scsi (target, scsi_cmd, 6, scsi_dat, 4);

  if (r != 0){
    scsi_cmd[0] = SC_RQS;
    scsi_cmd[4] = 4;
    scsi_cmd[1] = scsi_cmd[2] = 0;
    scsi_cmd[3] = scsi_cmd[5] = 0;
    r = do_scsi (target, scsi_cmd, 6, scsi_dat, 4);
    if (r != 0){
      return ERR;
    }
  }

  if ((scsi_dat[0] & 0x7e) == 0x70) {
	switch (scsi_dat[2] & 0xf) {
	case NoSense:
	case RecoveredError:
		r = OK;
		break;
	case UnitAttention:
	case 0x0b:	/* ABORT COMMAND */
		pi[target << 4].pi_size = 0;
		r = RETRY;
		break;
	default:
		r = ERR;
		break;
	}
  } else
	r = ERR;

  if (r == ERR) {
	printf ("sd%d: sense:", target);
	for (s = 0; s < 4; s++)
		printf (" %02x", scsi_dat[s]);
	printf ("\n");
  }

  return r;
}

PRIVATE int do_scsi (target, command, clen, data, dlen)
int target;
unsigned char *command;
int clen;
char *data;
int dlen;
{
  int b, s, m, i;
  int selected = 0;

  do {
	b = SDC->sc_psns & 0x9f;
	SDC->sc_pctl = b & 7;		/* keep message, c/d, i/o */
	switch (b) {
	case PS_BUS:			/* bus free phase */
		if (selected) {
			printf ("sd%d: unexpected phase %02x\n", target, b);
			return -1;
		}
		do {} while(SDC->sc_psns & 0x80);
		SDC->sc_ints = 0xff;
		s = scsi_selection (target);
		if (s != 0)
			return 0x80;
		selected = 1;
		break;
	case PS_CMD:			/* command phase */
		scsi_data_out (command, clen);
		break;
	case PS_DIN:			/* data in phase */
		scsi_data_in (data, dlen);
		break;
	case PS_DOT:			/* data out phase */
		scsi_data_out (data, dlen);
		break;
	case PS_STT:			/* status phase */
		s = scsi_byte_in ();
		break;
	case PS_MIN:			/* message in phase */
		m = scsi_byte_in ();
		if (m == 1) {
			i = scsi_byte_in ();
			m |= scsi_byte_in () << 8;
			while (--i > 0)
				scsi_byte_in ();
		}
		break;
	}
  } while (b != PS_MIN);		/* message in phase */

  if ((s & 0x1e) > 2)
	printf ("sd%d: status %02x message %02x\n", target, s, m);

  return s & 0x1e;
}

PRIVATE int scsi_selection (target)
int target;
{
  register int tm;
  register struct sdc *sdc;

  sdc = SDC;

  sdc->sc_pctl = 0;
  sdc->sc_temp = sdc->sc_bdid | (1 << target);
  sdc->sc_tch = timeout >> 8;
  sdc->sc_tcm = timeout;
  sdc->sc_tcl = 3;
  sdc->sc_scmd = 0x20;

  while (!(sdc->sc_ints & 0x14));
  sdc->sc_ints = 0x10;
  if (sdc->sc_ints & 4) {
	sdc->sc_ints = 0x04;
	return ERROR;
  }

  return OK;
}

PRIVATE void scsi_data_out (data, bytes)
unsigned char *data;
register int bytes;
{
  register struct dma *dma;
  register struct sdc *sdc;
  register unsigned char *sns, *dat, *cmd;

  dma = DMA1;
  sdc = SDC;
  cmd = &sdc->sc_scmd;

  sdc->sc_tch = 0;
  sdc->sc_tcm = bytes >> 8;
  sdc->sc_tcl = bytes;

#ifndef DONT_USE_DMA
  if (bytes < 512) {
#endif
	sns = &sdc->sc_psns;
	dat = &sdc->sc_temp;
	*cmd = 0xa4;				/* prog. transfer */
	do {
		do {} while (!(*sns & 0x80));	/* wait for req release */
		*dat = *data++;
		*cmd = 0xe4;			/* assert ack */
		do {} while (*sns & 0x80);	/* wait for req assert */
		*cmd = 0xc4;			/* release ack */
	} while (--bytes > 0);
#ifndef DONT_USE_DMA
  } else {
	dmagrab (WINCHESTER2);

	dma->dar = (long) &SDC->sc_dreg;
	dma->niv = 0x74;
	dma->eiv = 0x75;

	dma->mar = (long) data;			/* dma address setup */
	dma->mtc = bytes;

	do {
		dma->ocr = 0x30;
		*cmd = 0x80;			/* dma transfer */
		dma->ccr = 0x88;		/* start dma */
		receive(HARDWARE, &mess);	/* Wait for interrupt. */
	} while (!(sdc->sc_ints & 0x10));
	sdc->sc_ints = 0x10;

	dmafree (WINCHESTER2);
  }
#endif
}

PRIVATE void scsi_byte_out (data)
unsigned char data;
{
  register struct sdc *sdc;

  sdc = SDC;

  do {} while (!(sdc->sc_psns & 0x80));	/* wait for req release */
  sdc->sc_temp = data;
  sdc->sc_scmd = 0xe4;			/* assert ack */
  do {} while (sdc->sc_psns & 0x80);	/* wait for req assert */
  sdc->sc_scmd = 0xc4;			/* release ack */
}

PRIVATE void scsi_data_in (data, bytes)
register unsigned char *data;
register int bytes;
{
  register struct dma *dma;
  register struct sdc *sdc;
  register unsigned char *sns, *dat, *cmd;

  dma = DMA1;
  sdc = SDC;
  cmd = &sdc->sc_scmd;

  sdc->sc_tch = 0;
  sdc->sc_tcm = bytes >> 8;
  sdc->sc_tcl = bytes;

#ifndef DONT_USE_DMA
  if (bytes < 512) {
#endif
	sns = &sdc->sc_psns;
	dat = &sdc->sc_temp;
	*cmd = 0xa4;				/* prog. transfer */
	do {
		do {} while (!(*sns & 0x80));	/* wait for req release */
		*data++ = *dat;
		*cmd = 0xe4;			/* assert ack */
		do {} while (*sns & 0x80);	/* wait for req assert */
		*cmd = 0xc4;			/* release ack */
	} while (--bytes > 0);
#ifndef DONT_USE_DMA
  } else {
	dmagrab (WINCHESTER2);

	dma->dar = (long) &SDC->sc_dreg;
	dma->niv = 0x74;
	dma->eiv = 0x75;

	dma->mar = (long) data;			/* dma address setup */
	dma->mtc = bytes;

	do {
		dma->ocr = 0xb0;
		*cmd = 0x80;			/* dma transfer */
		dma->ccr = 0x88;		/* start dma */
		receive(HARDWARE, &mess);	/* Wait for interrupt. */
	} while (!(sdc->sc_ints & 0x10));
	sdc->sc_ints = 0x10;

	dmafree (WINCHESTER2);
  }
#endif
}

PRIVATE unsigned char scsi_byte_in ()
{
  unsigned char data;
  register struct sdc *sdc;

  sdc = SDC;

  do {} while (!(sdc->sc_psns & 0x80));	/* wait for req release */
  data = sdc->sc_temp;
  sdc->sc_scmd = 0xe4;			/* assert ack */
  do {} while (sdc->sc_psns & 0x80);	/* wait for req assert */
  sdc->sc_scmd = 0xc4;			/* release ack */

  return data;
}

#include <sgtty.h>

/*===========================================================================*
 *				do_ioctl				     * 
 *===========================================================================*/
PRIVATE int do_ioctl(mp)
message *mp;			/* pointer to read or write w_message */
{
  register	r, target;

  target = mp->DEVICE >> 4;
  if (target < 0 || target >= MAX_DRIVES)
	return(EIO);

  if (mp->COUNT == TIOCFLUSH)
	/*
	 * remake the partition description for the devices
	 */
		pi[target << 4].pi_size = 0;
  else
	return(EINVAL);

  return(r);
}

PRIVATE int get_p_info (target)
int target;
{
  register	r, minor;

  /* read sector 0 of the device and check initialized one */
  r = do_xfer(target, (phys_bytes) b, 0L, 1, DISK_READ);
  if (r != OK)
	return(EIO);
  if (strncmp (b, "X68SCSI1", 8)) {
	minor = (target << 4) + 1;
	for (r = 1; r < NPARTS; r++, minor++) {
		pi[minor].pi_name[0] = 0;
		pi[minor].pi_name[1] = 0;
		pi[minor].pi_start = 0;
		pi[minor].pi_size = 0;
	}
	return 0;
  }

  /* read sector 4 of the device and copy the partition info */
  r = do_xfer(target, (phys_bytes) hi, 4L, 1, DISK_READ);
  if (r != OK)
	return(EIO);
  minor = (target << 4) + 1;
  for (r = 1; r < NPARTS; r++, minor++) {
	if (*(char *) hi[r].hi_name) {
		pi[minor].pi_name[0] = hi[r].hi_name[0];
		pi[minor].pi_name[1] = hi[r].hi_name[1];
		pi[minor].pi_start = (hi[r].hi_start & 0xffffff) << 1;
		pi[minor].pi_size = hi[r].hi_size << 1;
	} else {
		pi[minor].pi_name[0] = 0;
		pi[minor].pi_name[1] = 0;
		pi[minor].pi_start = 0;
		pi[minor].pi_size = 0;
	}
  }

  return OK;
}

PRIVATE int get_d_info (target)
int target;
{
  register int	r, m, s;
  register unsigned char *p;

  if (target == HOST_ID)
	return ERR;
  if (SASI_FLAG (target))
	return ERR;

  m = target << 4;

  scsi_cmd[0] = SC_INQ;
  scsi_cmd[1] = scsi_cmd[2] = scsi_cmd[3] = scsi_cmd[5] = 0;
  scsi_cmd[4] = 3;

  timeout = 2500;				/* 500 ms */
  r = do_scsi (target, scsi_cmd, 6, pi[m].pi_name, 3);/* Inquiry */
  timeout = 25000;				/* 5000 ms */
  if (r != 0)
	return EIO;
  
  D_TYPE (target) &= 0x80;
  ANSI_V (target) &= 7;
  D_AVAL (target) = 1;
  if (D_TYPE (target) == 1) {		/* tape device */
	D_AVAL (target) = 0;
	return EIO;
  }

  scsi_cmd[0] = SC_CAP;
  scsi_cmd[1] = scsi_cmd[2] = scsi_cmd[3] = 0;
  scsi_cmd[4] = scsi_cmd[5] = scsi_cmd[6] = 0;
  scsi_cmd[7] = scsi_cmd[8] = scsi_cmd[9] = 0;

  timeout = 2500;				/* 500 ms */
  r = do_scsi (target, scsi_cmd, 10, scsi_dat, 8);	/* Read Capacity */
  timeout = 25000;				/* 5000 ms */
  if (r != 0)
	return EIO;

  pi[m].pi_name[1] = *(long *) (scsi_dat + 4);
  pi[m].pi_size = *(long *) (scsi_dat + 0) + 1;

  if (pi[m].pi_name[1] != 512) {
	scsi_cmd[0] = SC_MSL;
	scsi_cmd[1] = scsi_cmd[2] = scsi_cmd[3] = scsi_cmd[5] = 0;
	scsi_cmd[4] = 12;

	scsi_dat[3] = 8;
	scsi_dat[10] = SECTOR_SIZE >> 8;
	scsi_dat[1] = scsi_dat[2] = scsi_dat[4] = 0;
	scsi_dat[5] = scsi_dat[6] = scsi_dat[7] = 0;
	scsi_dat[8] = scsi_dat[9] = scsi_dat[11] = 0;

	timeout = 2500;				/* 500 ms */
	r = do_scsi (target, scsi_cmd, 6, scsi_dat, 12);	/* Mode Select */
	timeout = 25000;			/* 5000 ms */
	if (r != 0) {
		D_AVAL (target) = 0;
		return EIO;
	}

	pi[m].pi_size *= pi[m].pi_name[1] / SECTOR_SIZE;
	pi[m].pi_name[1] = SECTOR_SIZE;
  }

  return 0;
}

/*===========================================================================*
 *				sdcint					     *
 *===========================================================================*/

PUBLIC void sdcint()
{
  DMA1->csr = 0xff;
  interrupt(WINCHESTER2);
}

PUBLIC void sdeint()
{
  dma_s = DMA1->csr;
  dma_e = DMA1->cer;
  if (dma_s & 0x10 && dma_e == 0x0a) {
	SDC->sc_pctl = SDC->sc_psns & 7;	/* keep message, c/d, i/o */
	dma_s = dma_e = 0;
  } else {
	printf ("dma error interrupt accepted.\n");
	printf ("status %02x %02x.\n", dma_s, dma_e);
  }
  DMA1->csr = 0xff;
  interrupt(WINCHESTER2);
}
#endif 
